from mjooln import File, Atom


class LeafError(Exception):
    pass


class Leaf(File, Atom):
    """ Existing file within a tree that follows atom naming."""

    def __init__(self, file: File):
        File.__init__(self)
        if not self.exists():
            raise LeafError(f'Leaf path does not exist: {self}')
        if not Atom.is_seed(self.stub()):
            raise LeafError(f'File is not leaf: {self}')
        atom = Atom.from_seed(self.stub())
        Atom.__init__(self,
                      key=atom.key(),
                      identity=atom.identity(),
                      zulu=atom.zulu())



import uuid

from mjooln.core.seed import Seed


class IdentityError(Exception):
    pass


class Identity(str, Seed):
    """ UUID string generator with convenience functions

    Inherits str, and is therefore an immutable string, with a fixed format
    as illustrated below.

    Examples::

        Identity()
            'BD8E446D_3EB9_4396_8173_FA1CF146203C'

        Identity.is_in('Has BD8E446D_3EB9_4396_8173_FA1CF146203C within')
            True

        Identity.find_one('Has BD8E446D_3EB9_4396_8173_FA1CF146203C within')
            'BD8E446D_3EB9_4396_8173_FA1CF146203C'

    """

    REGEX = r'[0-9A-F]{8}\_[0-9A-F]{4}\_[0-9A-F]{4}\_[0-9A-F]{4}' \
             r'\_[0-9A-F]{12}'
    LENGTH = 36

    @classmethod
    def from_seed(cls, str_: str):
        return cls(str_)

    def __new__(cls,
                identity: str = None):
        if not identity:
            identity = str(uuid.uuid4()).replace('-', '_').upper()
        elif not cls.is_seed(identity):
            raise IdentityError(f'String is not valid identity: {identity}')
        instance = super(Identity, cls).__new__(cls, identity)
        return instance

    def __repr__(self):
        return f'Identity(\'{self}\')'

    @classmethod
    def elf(cls, identity):
        """ Returns new or existing identity

        None will create a new Identity, a string will be parsed as Identity,
        and an existing Identity will be returned as itself.

        :type identity: None, str or Identity
        :return: Identity
        """
        if isinstance(identity, Identity):
            return identity
        else:
            return cls(identity)



import os
import glob
import logging
import psutil
import socket
from sys import platform

from mjooln.atom.zulu import Zulu

logger = logging.getLogger(__name__)


class PathError(Exception):
    pass


# TODO: Handle network drives (i.e. not mounted)?
# TODO: Add the lack of speed in documentation. Not meant to be used for
# TODO: for large folders (thats the whole point)
class Path(str):
    """ Absolute paths as a string with convenience functions

    No relative paths are allowed. Paths not starting with a valid
    mountpoint will be based in current folder.

    All backslashes are replaced with forward slash.
    """

    FOLDER_SEPARATOR = '/'
    PATH_CHARACTER_LIMIT = 256

    LINUX = 'linux'
    WINDOWS = 'windows'
    OSX = 'osx'
    PLATFORM = {
        'linux': LINUX,
        'linux2': LINUX,
        'darwin': OSX,
        'win32': WINDOWS,
    }

    @classmethod
    def platform(cls):
        """ Get platform name

        :raises PathError: If platform is unknown
        :return: Platform name (linux/windows/osx)
        :rtype: str
        """
        if platform in cls.PLATFORM:
            return cls.PLATFORM[platform]
        else:
            raise PathError(f'Unknown platform {platform}. '
                            f'Known platforms are: {cls.PLATFORM.keys()}')

    @classmethod
    def host(cls):
        """ Get host name

        Wrapper for ``socket.gethostname()``

        :return: Host name
        :rtype: str
        """
        return socket.gethostname()

    @classmethod
    def join(cls, *args):
        """ Join strings to path

        Wrapper for ``os.path.join()``

        Relative paths will include current folder::

            Path.current()
                '/Users/zaphod/dev'
            Path.join('code', 'donald')
                '/Users/zaphod/dev/code/donald'

        :return: Absolute path
        :rtype: Path
        """
        return cls(os.path.join(*args))

    @classmethod
    def mountpoints(cls):
        """ List valid mountpoints/partitions or drives

        Finds mountpoints/partitions on linux/osx, and drives (C:, D:) on
        windows.

        :return: Valid mountpoints or drives
        :rtype: list
        """
        return [x.mountpoint.replace('\\', cls.FOLDER_SEPARATOR)
                for x in psutil.disk_partitions(all=True)]

    @classmethod
    def has_valid_mountpoint(cls, path_str):
        """ Flags if the path starts with a valid mountpoint

        :return: True if path has valid mountpoint, False if not
        :rtype: bool
        """
        return len([x for x in cls.mountpoints()
                    if path_str.startswith(x)]) > 0

    @classmethod
    def elf(cls, path, **kwargs):
        if isinstance(path, cls):
            return path
        else:
            return cls(path)

    def __new__(cls, path: str):
        if not isinstance(path, str):
            raise PathError(f'Input to constructor must be of type str')
        if not os.path.isabs(path):
            path = path.replace('\\', cls.FOLDER_SEPARATOR)
            path = os.path.abspath(path)
        path = path.replace('\\', cls.FOLDER_SEPARATOR)
        if len(path) > cls.PATH_CHARACTER_LIMIT:
            logger.warning(f'Path exceeds {cls.PATH_CHARACTER_LIMIT} '
                           f'characters, which may cause problems on some '
                           f'platforms')
        # TODO: Add check on valid names/characters
        self = super(Path, cls).__new__(cls, path)
        if not cls.has_valid_mountpoint(path):
            raise PathError(f'Path does not have valid mountpoint '
                            f'for this platform: {path}')
        return self

    # Make pandas (and other libraries) recognize Path class as pathlike
    def __fspath__(self):
        return str(self)

    def __repr__(self):
        return f'Path(\'{self}\')'

    def name(self):
        """ Get name of folder or file

        Example::

            f = Folder.current()
            f
                '/Users/zaphod'
            f.name()
                'donald'

            f = f.join('dev', 'code', 'donald')
            f
                '/Users/zaphod/dev/code/donald'
            f.name()
                'donald'

        :return: Name of folder or file
        :rtype: str
        """
        return os.path.basename(self)

    def volume(self):
        """ Return path volume

        Volume is a collective term for mountpoint or drive

        :raises PathError: If volume cannot be determined
        :return: Volume of path
        :rtype: Path
        """
        mountpoints = self.mountpoints()
        candidates = [x for x in mountpoints if self.startswith(x)]
        if len(candidates) > 1:
            candidates = [x for x in candidates
                          if not x == self.FOLDER_SEPARATOR]
        if len(candidates) == 1:
            return Path(candidates[0])
        else:
            raise PathError(f'Could not determine volume: {mountpoints}')

    def exists(self):
        """ Check if path exists

        Wrapper for ``os.path.exists()``

        :return: True if path exists, False otherwise
        :rtype: bool
        """
        return os.path.exists(self)

    def raise_if_not_exists(self):
        """ Raises an exception if path does not exist

        :raises PathError: If path does not exist
        """
        if not self.exists():
            raise PathError(f'Path does not exist: {self}')

    def is_volume(self):
        # TODO: Add handling of network drive
        # TODO: Check if found as network drive if not in mountpoints
        """ Check if path is a volume

        :raises PathError: If path does not exist
        :return: True if path is a volume, False if not
        :rtype: bool
        """
        if self.exists():
            return self in self.mountpoints()
        else:
            raise PathError(f'Cannot see if non existent path '
                            f'is a volume or not: {self}')

    def is_folder(self):
        """ Check if path is a folder

        :raises PathError: If path does not exist
        :return: True if path is a folder, False if not
        :rtype: bool
        """
        if self.exists():
            return os.path.isdir(self)
        else:
            raise PathError(f'Cannot determine if non existent path '
                            f'is a folder or not: {self}')

    def is_file(self):
        """ Check if path is a file

        :raises PathError: If path does not exist
        :return: True if path is a file, False if not
        :rtype: bool
        """
        if self.exists():
            return os.path.isfile(self)
        else:
            raise PathError(f'Cannot determine if non existent path '
                            f'is a file or not: {self}')

    def size(self):
        """ Return file or folder size

        .. warning:: If Path is a folder, ``size()`` will return a small number,
            representing the size of the folder object, not its contents.
            For finding actual disk usage of a folder, use
            :meth:`.Folder.disk_usage()`

        :raises PathError: If path does not exist
        :returns: File or folder size
        :rtype: int
        """
        if self.exists():
            return os.stat(self).st_size
        else:
            raise PathError(f'Cannot determine size of '
                            f'non existent path: {self}')

    def created(self):
        """ Get created timestamp

        :return: Timestamp created
        :rtype: Zulu
        """
        return Zulu.from_epoch(os.stat(self).st_ctime)

    def modified(self):
        """
        Get modified timestamp

        :returns: Timestamp modified
        :rtype: Zulu
        """
        return Zulu.from_epoch(os.stat(self).st_mtime)

    def parts(self):
        """ Get list of parts in path

        :returns: String parts of path
        :rtype: list
        """
        parts = str(self).split(self.FOLDER_SEPARATOR)
        # Remove empty first part (if path starts with /)
        if parts[0] == '':
            parts = parts[1:]
        return parts



import gzip
import logging
import os
import shutil
import hashlib
from pathlib import Path as Path_

from mjooln.core.crypt import Crypt
from mjooln.core.doc import Doc, JSON, YAML
from mjooln.filesystem.path import Path
from mjooln.filesystem.folder import Folder

logger = logging.getLogger(__name__)


# TODO: Add yaml files


class FileError(Exception):
    pass


class FileName(str):
    JSON = 'json'
    YAML = 'yaml'
    TEXT = 'text'

    TYPES = [JSON, YAML, TEXT]

    MODE_TEXT = 't'
    MODE_BINARY = 'b'

    TYPE = {
        'json': {
            'extensions': ['json'],
            'mode': MODE_TEXT,
        },
        'yaml': {
            'extensions': ['yml', 'yaml'],
            'mode': MODE_TEXT,
        },
        'text': {
            'extensions': ['log', 'txt'],
            'mode': MODE_TEXT,
        }
    }

    #: Files with this extension will compress text before writing to file
    #: and decompress after reading
    COMPRESSED_EXTENSION = 'gz'

    #: Files with this extension will encrypt before writing to file, and
    #: decrypt after reading. The read/write methods therefore require a
    #: crypt_key
    ENCRYPTED_EXTENSION = 'aes'

    # #: Extensions reserved for compression and encryption
    # RESERVED_EXTENSIONS = [COMPRESSED_EXTENSION,
    #                        ENCRYPTED_EXTENSION]

    #: File names starting with this character will be tagged as hidden
    HIDDEN_STARTSWITH = '.'

    #: Extension separator. Period
    EXTENSION_SEPARATOR = '.'

    @classmethod
    def get_type(cls, extension):
        for key, item in cls.TYPE.items():
            if extension in item['extensions']:
                return key
        return None

    @classmethod
    def make(cls,
             stub: str,
             extension: str,
             is_hidden: bool = False,
             is_compressed: bool = False,
             is_encrypted: bool = False):
        """
        Create file name from stub and attributes

        :param stub: File stub, barring extensions and hidden startswith
        :param extension: File extension
        :param is_hidden: True tags file name as hidden
        :param is_compressed: True tags file name as compressed, adding the
            necessary extra extension
        :param is_encrypted:
        :return:
        """
        if stub.startswith(cls.HIDDEN_STARTSWITH):
            raise FileError(f'Stub cannot start with the hidden flag. '
                            f'Keep stub clean, and set is_hidden=True')
        if cls.EXTENSION_SEPARATOR in stub:
            raise FileError(f'Cannot add stub with extension '
                            f'separator in it: {stub}. '
                            f'Need a clean string for this')
        if cls.EXTENSION_SEPARATOR in extension:
            raise FileError(f'Cannot add extension with extension '
                            f'separator in it: {extension}. '
                            f'Need a clean string for this')
        names = [stub, extension]
        if is_compressed:
            names += [cls.COMPRESSED_EXTENSION]
        if is_encrypted:
            names += [cls.ENCRYPTED_EXTENSION]
        name = cls.EXTENSION_SEPARATOR.join(names)
        if is_hidden:
            name = cls.HIDDEN_STARTSWITH + name
        return cls(name)

    def __new__(cls, name):
        self = super(FileName, cls).__new__(cls, name)
        self.hidden = name.startswith(self.HIDDEN_STARTSWITH)
        if self.hidden:
            name = name[1:]
        parts = name.split('.')
        while not parts[0]:
            parts = parts[1:]
        self.stub = parts[0]
        self.extensions = parts[1:]
        self.encrypted = self.extensions \
                         and self.extensions[-1] == cls.ENCRYPTED_EXTENSION
        if self.encrypted:
            self.extensions = self.extensions[:-1]
        self.compressed = self.extensions \
                          and self.extensions[-1] == cls.COMPRESSED_EXTENSION
        if self.compressed:
            self.extensions = self.extensions[:-1]
        self.extension = None
        if len(self.extensions) == 1:
            self.extension = self.extensions[0]
        self.type = None
        self.mode = ''
        if self.extension:
            self.type = cls.get_type(self.extension)
        if self.type:
            self.mode = cls.TYPE[self.type]['mode']
        return self

    def __repr__(self):
        return f'FileName(\'{self}\')'

    # def __init__(self, name: str):
    #     super(FileName, self).__init__(name)

    # def parts(self):


class File(Path):
    """
    Convenience class for file handling

    Create and read a file::

        f = File('my_file.txt')
        f.write('Hello world')
        f.read()
            'Hello world'
        f.size()
            11

    Compress and encrypt::

        fc = f.compress()
        fc.name()
            'my_file.txt.gz'
        fc.read()
            'Hello world'

        crypt_key = Crypt.generate_key()
        crypt_key
            b'aLQYOIxZOLllYThEKoXTH_eqTQGEnXm9CUl2glq3a2M='
        fe = fc.encrypt(crypt_key)
        fe.name()
            'my_file.txt.gz.aes'
        fe.read(crypt_key=crypt_key)
            'Hello world'

    Create an encrypted file, and write to it::

        ff = File('my_special_file.txt.aes')
        ff.write('Hello there', password='123')
        ff.read(password='123')
            'Hello there'

        f = open(ff)
        f.read()
            'gAAAAABe0BYqPPYfzha3AKNyQCorg4TT8DcJ4XxtYhMs7ksx22GiVC03WcrMTnvJLjTLNYCz_N6OCmSVwk29Q9hoQ-UkN0Sbbg=='
        f.close()

    .. note:: Using the ``password`` parameter, builds an encryption key by
        combining it with the builtin (i.e. hard coded) class salt.
        For proper security, generate your
        own salt with :meth:`.Crypt.salt()`. Then use
        :meth:`.Crypt.key_from_password()` to generate a crypt_key

    .. warning:: \'123\' is not a real password

    """
    _salt = b'O89ogfFYLGUts3BM1dat4vcQ'


    # TODO: Add binary flag based on extension (all other than text is binary..)
    # TODO: Facilitate child classes with custom read/write needs

    @classmethod
    def join(cls, *args):
        # Purely cosmetic for IDE
        return super().join(*args)

    @classmethod
    def home(cls,
             file_name: str):
        """
        Create a file path in home folder

        :param file_name: File name
        :type file_name: str
        :return: File
        :rtype: File
        """
        return cls.join(Folder.home(), file_name)

    @classmethod
    def _crypt_key(cls,
                   crypt_key: bytes = None,
                   password: str = None):
        """ Using a password will make a encryption_key combined with the
        internal class salt
        """
        if crypt_key and password:
            raise FileError('Use either crypt_key or password.')
        elif not crypt_key and not password:
            raise FileError('crypt_key or password missing')
        if crypt_key:
            return crypt_key
        else:
            return Crypt.key_from_password(cls._salt, password)

    def __new__(cls,
                path: str,
                is_binary: bool = False,
                *args, **kwargs):
        # TODO: Raise exception if reserved extensions are used inappropriately
        self = Path.__new__(cls, path)
        if self.exists():
            if self.is_volume():
                raise FileError(f'Path is volume, not file: {path}')
            elif self.is_folder():
                raise FileError(f'Path is existing folder, '
                                f'not file: {path}')

        self._name = FileName(self.name())
        return self

    # TODO: Add touch (and untouch?)
    def __repr__(self):
        return f'File(\'{self}\')'

    def parts(self):
        """
        Get file parts, i.e. those separated by period

        :return: list
        """
        return self._name.parts

    def touch(self):
        Path_(self).touch()

    def untouch(self, ignore_if_not_empty=False):
        if self.exists():
            if self.size() == 0:
                self.delete()
            else:
                if not ignore_if_not_empty:
                    raise FileError(f'Cannot untouch file '
                                    f'that is not empty: {self}')

    def extensions(self):
        """
        Get file extensions

        :return: List of file extensions
        :rtype: list
        """
        return self._name.extensions

    def is_hidden(self):
        """
        Check if file is hidden, i.e. starts with a period

        :return: True if hidden, False if not
        :rtype: bool
        """
        return self._name.hidden

    def is_encrypted(self):
        """
        Check if file is encrypted, i.e. ends with reserved extension \'aes\'

        :return: True if encrypted, False if not
        :rtype: bool
        """
        return self._name.encrypted

    def is_compressed(self):
        """
        Check if file is compressed, i.e. has reserved extension \'gz\'

        :return: True if compressed, False if not
        :rtype: bool
        """
        return self._name.compressed

    def stub(self):
        """
        Get file stub, i.e. the part of the file name bar extensions

        :return: File stub
        :rtype: str
        """
        return self._name.stub

    def extension(self):
        """
        Get file extension, i.e. the extension which is not reserved.
        A file is only supposed to have one extension that does not indicate
        either compression or encryption.

        :raise FileError: If file has more than one not reserved extensions
        :return: File extension
        :rtype: str
        """
        return self._name.extension

    def type(self):
        return self._name.type

    def mode(self):
        return self._name.mode

    def set_mode_binary(self):
        self._name.mode = self._name.MODE_BINARY

    def set_mode_text(self):
        self._name.mode = self._name.MODE_TEXT

    def md5_checksum(self):
        """
        Get MD5 Checksum for the file

        :raise FileError: If file does not exist
        :return: MD5 Checksum
        :rtype: str
        """
        if not self.exists():
            raise FileError(f'Cannot make checksum '
                            f'if file does not exist: {self}')
        md5 = hashlib.md5()
        with open(self, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5.update(chunk)
        return md5.hexdigest()

    def delete(self,
               missing_ok: bool = False):
        """
        Delete file

        :raise FileError: If file is missing, and ``missing_ok=False``
        :param missing_ok: Indicate if an exception should be raised if the
            file is missing. If True, an exception will not be raised
        :type missing_ok: bool
        :return: None
        """
        if self.exists():
            logger.debug(f'Delete file: {self}')
            os.unlink(self)
        elif not missing_ok:
            raise FileError(f'Tried to delete file '
                            f'that doesn\'t exist: {self}')

    def write(self, data, mode='w',
              crypt_key: bytes = None,
              password: str = None,
              # TODO: Handle JSON vs YAML write. I.e. maybe remove direct write
              human_readable: bool = True,
              **kwargs):
        """
        Write data to file

        For encryption, use either ``crypt_key`` or ``password``. None or both
        will raise an exception. Encryption requires the file name to end with
        extension \'aes\'

        :raise FileError: If using ``crypt_key`` or ``password``, and the
            file does not have encrypted extension
        :param data: Data to write
        :type data: str or bytes
        :param mode: Write mode
        :type mode: str
        :param crypt_key: Encryption key
        :type crypt_key: bytes
        :param password: Password (will use class salt)
        :type password: str
        :param human_readable: JSON only, will write json to file with
            new lines and indentation
        :type human_readable: bool
        """
        # TODO: Require compression if file is encrypted?
        if self.is_encrypted():
            crypt_key = self._crypt_key(crypt_key, password)
        elif crypt_key or password:
            raise FileError(f'File does not have crypt extension '
                            f'({self._name.ENCRYPTED_EXTENSION}), '
                            f'but a crypt_key '
                            f'or password was sent as input to write.')
        if self.is_json():
            if isinstance(data, Doc):
                data = data.to_json(human_readable=human_readable)
            elif isinstance(data, dict):
                data = JSON.dumps(data)
        elif self.is_yaml():
            if isinstance(data, Doc):
                data = data.to_yaml()
            elif isinstance(data, dict):
                data = YAML.dump(data)

        self.folder().touch()
        if self.is_compressed():
            self._write_compressed(data)
            if self.is_encrypted():
                if self.size() > 100000:
                    logger.warning('On the fly encrypt/compress not '
                                   'implemented. There is an extra read/write '
                                   'from/to disk. In other words, '
                                   'this is a hack.')
                # TODO: Refactor to write once, but verify zlib/gzip
                #  compatibility
                data = self._read(mode='rb')
                data = Crypt.encrypt(data, crypt_key)
                self.delete()
                self._write(data, mode='wb')
        else:
            if self.is_encrypted():
                if not isinstance(data, bytes):
                    data = data.encode()
                data = Crypt.encrypt(data, crypt_key)
                self._write(data, mode='wb')
            else:
                if mode and self.mode():
                    mode = mode + self.mode()
                self._write(data, mode=mode)

    def _write(self, data, mode='w'):
        with open(self, mode=mode) as f:
            f.write(data)

    def _write_compressed(self, content):
        if not isinstance(content, bytes):
            content = content.encode()
        with gzip.open(self, mode='wb') as f:
            f.write(content)

    def open(self, mode='r'):
        return open(self, mode=mode)

    def is_text(self):
        return self._name.type == self._name.TEXT

    def is_json(self):
        return self._name.type == self._name.JSON

    def is_yaml(self):
        return self._name.type == self._name.YAML

    def read(self, mode='r',
             crypt_key: bytes = None,
             password: str = None,
             num_lines: int = None,
             *args, **kwargs):
        """
        Read file

        If file is encrypted, use either ``crypt_key`` or ``password``.
        None or both will raise an exception. Encryption requires the file
        name to end with extension \'aes\'

        :raise FileError: If trying to decrypt a file without encryption
            extension
        :param mode: Read mode
        :param crypt_key: Encryption key
        :type crypt_key: bytes
        :param password: Password (will use class salt)
        :type password: str
        :param num_lines: Number of lines to read. ``None`` means read entire
            file (normal read), and this is also default
        :type num_lines: int
        :return: data
        :rtype: str or bytes
        """
        if not self.exists():
            raise FileError(f'Cannot read from file '
                            f'that does not exist: {self}')
        if self.is_encrypted():
            crypt_key = self._crypt_key(crypt_key, password)
        elif crypt_key or password:
            raise FileError(f'File does not have crypt extension '
                            f'({self._name.ENCRYPTED_EXTENSION}), '
                            f'but a crypt_key '
                            f'or password was sent as input to write.')
        if self.is_compressed():
            if self.is_encrypted():
                logger.warning('On the fly encrypt/compress not implemented. '
                               'There is an extra read/write from/to disk. '
                               'In other words, this is a hack.')
                # TODO: Refactor to read once, but
                #  verify zlib/gzip compatibility
                decrypted_file = self.decrypt(crypt_key, delete_original=False)
                data = decrypted_file._read_compressed(mode=mode)
                decrypted_file.delete()
            else:
                data = self._read_compressed(mode=mode, num_lines=num_lines)
        else:
            if self.is_encrypted():
                if num_lines is not None:
                    raise FileError(f'Read lines from encrypted file not '
                                    f'supported')
                data = self._read(mode='rb')
                data = Crypt.decrypt(data, crypt_key)
                if 'b' not in mode:
                    data = data.decode()
            else:
                mode = mode + self.mode()
                data = self._read(mode=mode,
                                  num_lines=num_lines)

        if self.is_json():
            if num_lines is not None:
                raise FileError(f'Cannot read lines from a json file: {self}')
            data = JSON.loads(data)
        elif self.is_yaml():
            if num_lines is not None:
                raise FileError(f'Cannot read lines from a yaml file: {self}')
            data = YAML.safe_load(data)

        return data

    def _read(self,
              mode='r',
              num_lines: int = None):
        with open(self, mode=mode) as f:
            if not num_lines:
                content = f.read()
            else:
                content = []
                for n in range(num_lines):
                    content.append(f.readline().strip())
                if len(content) == 1:
                    content = content[0]
        return content

    def _read_compressed(self,
                         mode='rb',
                         num_lines: int = None):
        with gzip.open(self, mode=mode) as f:
            if not num_lines:
                content = f.read()
            else:
                content = []
                for n in range(num_lines):
                    content.append(f.readline().strip())
                if len(content) == 1:
                    content = content[0]
        if 'b' not in mode:
            if not isinstance(content, str):
                if isinstance(content, list):
                    for n in range(len(content)):
                        if isinstance(content[n], bytes):
                            content[n] = content[n].decode()
                elif isinstance(content, bytes):
                    content = content.decode()

        return content

    def rename(self,
               new_name: str):
        """
        Rename file

        :param new_name: New file name, including extension
        :type new_name: str
        :return: A new File with the new file name
        :rtype: File
        """
        new_path = self.join(self.folder(), new_name)
        os.rename(self, new_path)
        return File(new_path)

    def folder(self):
        """
        Get the folder containing the file

        :return: Folder containing the file
        :rtype: Folder
        """
        return Folder(os.path.dirname(self))

    # def files(self,
    #           pattern: str = '*',
    #           recursive: bool = False):
    #     """
    #     Get a list of all files in folder containing the file
    #
    #     :param pattern: Wildcard or pattern
    #     :param recursive: Recursive flag. When True, all subfolders will be
    #         searched
    #     :return: List of files
    #     :rtype: list of File
    #     """
    #     files = self.folder().list(pattern='*', recursive=False)
    #     return [File(x) for x in files if x.is_file()]

    def move(self,
             new_folder: Folder,
             new_name: bool = None):
        """
        Move file to a new folder, and optionally a new name

        :param new_folder: New folder
        :type new_folder: Folder
        :param new_name: New file name (optional). If missing, the file will
            keep the same name
        :return: Moved file
        :rtype: File
        """
        new_folder.touch()
        if new_name:
            new_file = File.join(new_folder, new_name)
        else:
            new_file = File.join(new_folder, self.name())
        if self.volume() == new_folder.volume():
            os.rename(self, new_file)
        else:
            shutil.move(self, new_file)
        return new_file

    def copy(self,
             new_folder: Folder,
             new_name: bool = None):
        """
        Copy file to a new folder, and optionally give it a new name

        :param new_folder: New folder
        :type new_folder: Folder
        :param new_name: New file name (optional). If missing, the file will
            keep the same name
        :return: Copied file
        :rtype: File
        """
        if self.folder() == new_folder:
            raise FileError(f'Cannot copy a file '
                            f'to the same folder: {new_folder}')
        new_folder.touch()
        if new_name:
            new_file = File.join(new_folder, new_name)
        else:
            new_file = File.join(new_folder, self.name())
        shutil.copyfile(self, new_file)
        return new_file

    def compress(self,
                 delete_original: bool = True):
        """
        Compress file

        :param delete_original: If True, original file will be deleted after
            compression
        :return: New compressed file, with extension \'gz\'
        :rtype: File
        """
        if self.is_compressed():
            raise FileError(f'File already compressed: {self}')
        if self.is_encrypted():
            raise FileError(f'Cannot compress encrypted file: {self}. '
                            f'Decrypt file first')

        logger.debug(f'Compress file: {self}')
        old_size = self.size()
        new_file = File(f'{self}.gz')
        if new_file.exists():
            logger.warning(f'Overwrite existing gz-file: {new_file}')
        with open(self, 'rb') as f_in:
            with gzip.open(str(new_file), 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        if delete_original:
            self.delete()
        new_file.compression_percent = 100 * (old_size - new_file.size()) \
                                       / old_size
        logger.debug(f'Compressed with compression '
                     f'{new_file.compression_percent:.2f}')
        return new_file

    def decompress(self,
                   delete_original: bool = True,
                   replace_if_exists: bool = True):
        """
        Decompress file

        :param delete_original: If True, the original compressed file will be
            deleted after decompression
        :param replace_if_exists: If True, the decompressed file will replace
            any already existing file with the same name
        :return: Decompressed file
        :rtype: File
        """
        if not self.is_compressed():
            raise FileError(f'File is not compressed: {self}')
        if self.is_encrypted():
            raise FileError(f'Cannot decompress encrypted file: {self}. '
                            f'Decrypt file first.')
        logger.debug(f'Decompress file: {self}')
        new_file = File(str(self).replace('.' + self._name.COMPRESSED_EXTENSION, ''))
        if new_file.exists():
            if replace_if_exists:
                logger.debug('Overwrite existing file: {}'.format(new_file))
            else:
                raise FileError(f'File already exists: \'{new_file}\'. '
                                f'Use replace_if_exists=True to ignore.')
        with gzip.open(self, 'rb') as f_in:
            with open(new_file, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        if delete_original:
            self.delete()
        new_file = File(new_file)
        new_file.compression_percent = None
        return new_file

    def encrypt(self,
                crypt_key: bytes,
                delete_original: bool = True):
        """
        Encrypt file

        :raise FileError: If file is already encrypted
        :param crypt_key: Encryption key
        :param delete_original: If True, the original unencrypted file will
            be deleted after encryption
        :return: Encrypted file
        :rtype: File
        """
        if self.is_encrypted():
            raise FileError(f'File is already encrypted: {self}')
        logger.debug(f'Encrypt file: {self}')
        encrypted_file = File(self + '.' + self._name.ENCRYPTED_EXTENSION)
        data = self._read(mode='rb')
        encrypted = Crypt.encrypt(data, crypt_key)
        encrypted_file._write(encrypted, mode='wb')
        if delete_original:
            self.delete()
        return encrypted_file

    def decrypt(self,
                crypt_key: bytes,
                delete_original: bool = True):
        """
        Decrypt file

        :raise FileError: If file is not encrypted
        :param crypt_key: Encryption key
        :param delete_original: If True, the original encrypted file will
            be deleted after decryption
        :return: Decrypted file
        :rtype: File
        """
        if not self.is_encrypted():
            raise FileError(f'File is not encrypted: {self}')

        logger.debug(f'Decrypt file: {self}')
        decrypted_file = File(self.replace('.' +
                                           self._name.ENCRYPTED_EXTENSION, ''))
        data = self._read(mode='rb')
        decrypted = Crypt.decrypt(data, crypt_key)
        decrypted_file._write(decrypted, mode='wb')
        if delete_original:
            self.delete()
        return decrypted_file



import logging

from mjooln import CryptError, Folder, File, Doc, Atom, AtomError, Element, \
    FileSystem, Zulu, FileName
from mjooln.tree.root import Root, NotRootException
from mjooln.tree.leaf import Leaf, LeafError

logger = logging.getLogger(__name__)


class TreeError(Exception):
    pass


class NotTreeException(NotRootException):
    pass


class Tree(Root):
    """
    Folder structure based on strict :class:`.Atom` file names, and with an
    autonomous vocabulary.

    Create a new Tree by planting it in any folder, and with a key following
    the limitations of :class:`Key`::

        folder = Folder.home()
        key = 'oak'
        oak = Tree.plant(folder, key)

    Files outside the tree are defined as ``native``, while files in the
    tree are called ``Leaf``. Leaves are uniquely identified by
    :class:`.Atom`, which also defines their position in the folder
    structure underlying the tree.

    Leaves can be grown from a native file by defining a corresponding
    atom::

        native = File('some_outside_file.csv')
        atom = Atom(key='my_key')
        oak.grow(native, atom)

    The file may later be retrieved

    """

    #: Tree class identifier
    TREE = Element('tree')

    #: Current class identifier (this is a Tree)
    SPECIES = TREE

    class Trunk(Doc):

        def __init__(self,
                     extension='csv',
                     compress_all=False,
                     encrypt_all=False,
                     key_level=None,
                     date_level=0,
                     time_level=0):
            self.extension = extension
            self.compress_all = compress_all
            self.encrypt_all = encrypt_all
            self.key_level = key_level
            self.date_level = date_level
            self.time_level = time_level

    TRUNK = Trunk

    @classmethod
    def is_tree(cls, folder):
        """
        Check if folder is a tree

        :param folder: Folder to check
        :type folder: Folder
        :return: True if folder is a tree, False if not
        :rtype: bool
        """
        try:
            _ = cls(folder)
            return True
        except NotRootException:
            return False

    #
    # @classmethod
    # def plant(cls,
    #           folder: Folder,
    #           key: Key,
    #           trunk: Trunk = Trunk()):
    #     return super(Tree, cls).plant(folder=folder,
    #                                   key=key,
    #                                   trunk=trunk)

    #     # TODO: Add warning for long keys that may result in long path names
    #     # zulu: 23, id: 36, seps: 2 x 3, max_path:??
    #     """
    #     Creates a new tree
    #
    #     Folder specifies where the tree should be planted, while key will
    #     be the name of the tree folder
    #
    #     :param folder: Folder to plant the tree in
    #     :type folder: Folder
    #     :param key: Tree name or key
    #     :type key: Key
    #     :param extension: File extension for all files/leaves in tree
    #     :type extension: str
    #     :param compress_all: Flags compression of all files/leaves in tree
    #     :type compress_all: bool
    #     :param encrypt_all: Flags encryption of all files/leaves in tree
    #     :type encrypt_all: bool
    #     :param key_level: Key levels for folder structure
    #     :type key_level: int/None
    #     :param date_level: Date levels for folder structure
    #     :type date_level: int/None
    #     :param time_level: Time levels for folder structure
    #     :type time_level: int/None
    #     :param encryption_key: (*optional*) Encryption key used for
    #         encryption/decription of files/leaves in tree
    #     :type encryption_key: bytes
    #     :param kwargs: Additional attributes of tree
    #     :return: New tree
    #     :rtype: Tree
    #     """
    #     return = super(Tree, cls).plant(folder, key, trunk)
    #     if trunk._encryption_key:
    #         tree.add_encryption_key(encryption_key)
    #     return tree

    def __init__(self, folder, encryption_key=None):
        super(Tree, self).__init__(folder)
        self._encryption_key = None
        if self.species != self.SPECIES:
            raise NotTreeException(f'Species mismatch. '
                                   f'Expected: {self.SPECIES}, '
                                   f'Found: {self.species}')
        if encryption_key:
            self.add_encryption_key(encryption_key)

    def __repr__(self):
        return f'Tree(\'{self._folder}\')'

    def add_encryption_key(self, encryption_key=None):
        """
        Adds encryption key to trees with encryption activated. Encryption key
        may also be added in constructor. It will not be stored on disk, only
        in memory.

        :raise TreeError: If encryption is activated, but encryption_key is
            None
        :param encryption_key: Encryption key as defined by :class:`.Crypt`
        :type encryption_key: bytes
        """
        if self.trunk.encrypt_all and encryption_key is None:
            raise TreeError(f'Tree is encrypt_all, but no encryption_key '
                            f'supplied in constructor. Check Crypt class on '
                            f'how to make one, or disable encryption')
        self._encryption_key = encryption_key

    def _branch(self, atom):
        """
        Get the folder for the given atom

        :param atom: Input atom
        :type atom: Atom
        :return: Folder where a leaf with the input atom would be
        :rtype: Folder
        """
        levels = atom.levels(key_level=self.trunk.key_level,
                             date_level=self.trunk.date_level,
                             time_level=self.trunk.time_level)
        return self._folder.append(levels)

    def grow(self,
             native,
             atom=None,
             force_extension=False,
             delete_native=False):
        """
        Add a file to the tree. A file outside of the tree is called ``native``
        while a file in the tree is called ``leaf``

        :param native: File to add to the tree
        :type native: File
        :param atom: Atom defining the contents of the file, if the file name
            is not a valid atom string
        :type atom: Atom
        :param force_extension: Flags forcing file extension to be the
            extension inherent in tree
        :type force_extension: bool
        :param delete_native: Flags whether to delete the native file after
            file has been copied to the tree structure
        :type delete_native: bool
        :return: New leaf in tree
        :rtype: Leaf
        """
        native = File.elf(native)
        if not native.exists():
            raise TreeError(f'Cannot grow Leaf from non existent '
                            f'path: {native}')
        if not native.is_file():
            raise TreeError(f'Native is not a file: {native}')
        if not force_extension and len(native.extensions()) != 1:
            raise TreeError(f'Leaves must have one extension, this native '
                            f'has none or multiple: {native.extensions()}. '
                            f'Use force_extension to set it to '
                            f'tree inherent: \'{self.trunk.extension}\'')

        if not atom:
            try:
                atom = Atom.from_seed(native.stub())
            except AtomError as se:
                raise TreeError(f'File name is not a valid '
                                f'atom: {native.name()}. '
                                f'Add atom as parameter to override '
                                f'file name.') from se

        if force_extension:
            extension = self.trunk.extension
        else:
            extension = native.extension()

        new_name = FileName.make(str(atom),
                                 extension,
                                 is_compressed=native.is_compressed(),
                                 is_encrypted=native.is_encrypted())

        if native.name() == new_name:
            new_name = None

        folder = self._branch(atom)
        if delete_native:
            file = native.move(folder, new_name)
        else:
            file = native.copy(folder, new_name)
        leaf = Leaf(file)
        logger.debug(f'Added leaf: {leaf}')
        leaf = self._shape(leaf)
        return leaf

    def leaf(self, atom):
        """
        Get the leaf defined by the given atom

        :param atom: Atom defining the leaf contents
        :type atom: Atom
        :return: Leaf defined by the atom
        :rtype: Leaf
        """
        atom = Atom.elf(atom)
        folder = self._branch(atom)
        file_name = FileName.make(str(atom),
                                  self.trunk.extension,
                                  is_compressed=self.trunk.compress_all,
                                  is_encrypted=self.trunk.encrypt_all)
        file = File.join(folder, file_name)
        return Leaf(file)

    def leaves(self, ignore_weeds: bool = False):
        # TODO: Rewrite to use FileSystem.walk()
        """
        Walk through all leaves in tree

        :return: Generator of leaves in tree
        :rtype: (Leaf)
        """

        for file in self.files():
            try:
                yield Leaf(file)
            except (LeafError, AtomError, ValueError) as e:
                if not ignore_weeds:
                    raise TreeError(f'File is not leaf: {e}; '
                                    f'Use prune() to remove weeds (not leaf), '
                                    f'or set ignore_weeds=True')

    def _shape(self, leaf):
        """
        Performs compress/decompress and/or encrypt/decrypt on a given file
        to match tree settings

        :param leaf: Leaf to shape
        :type leaf: Leaf
        :return: Shaped leaf
        :rtype: Leaf
        """
        try:
            if not leaf.is_compressed() and self.trunk.compress_all:
                if leaf.is_encrypted():
                    leaf = leaf.decrypt(self._encryption_key)
                leaf = leaf.compress()
            elif leaf.is_compressed() and not self.trunk.compress_all:
                if leaf.is_encrypted():
                    leaf = leaf.decrypt(self._encryption_key)
                leaf = leaf.decompress()
            if not leaf.is_encrypted() and self.trunk.encrypt_all:
                leaf = leaf.encrypt(self._encryption_key)
            elif leaf.is_encrypted() and not self.trunk.encrypt_all:
                leaf = leaf.decrypt(self._encryption_key)
        except CryptError as ce:
            raise TreeError(f'Invalid or missing encryption key while '
                            f'attempting reshape of {leaf}. '
                            f'Original error: {ce}')

        return Leaf.elf(leaf)

    def reshape(self, **kwargs):
        """
        Changes folder structure and/or compress/encrypt settings for all
        files in the tree from the input keyword arguments

        Performs compress/decompress and/or encrypt/decrypt on all files in
        tree to match tree settings for is_compressed and is_encrypted.

        Also prunes the tree moving files to correct folders, and deleting
        unused folders according to level settings

        .. note:: Folder and key cannot be changed

        :raise TreeError: If kwargs contain ``folder`` or ``key``
        :param kwargs: Attributes to change. Allowed attributes as in
            :meth:`plant()`
        """
        if not kwargs:
            logger.debug('No input arguments. Skipping reshape.')
        else:
            if 'folder' in kwargs:
                raise TreeError('Cannot change the folder of a tree '
                                'with reshape')
            if 'key' in kwargs:
                raise TreeError('Cannot change the key of a tree with '
                                'reshape')
            self.trunk.add(kwargs)
            self.touch()
            self.prune()

    def _prune_leaf(self, leaf):
        if not leaf.folder() == self._branch(leaf):
            leaf = leaf.move(self._branch(leaf))
            logger.debug(f'Move leaf: {leaf}')
        return leaf

    def prune(self, delete_not_leaf=False):
        """
        Moves files and deletes unused folders so that the tree structure
        matches level settings

        :param delete_not_leaf: Whether to delete files that are not leaf
        :type delete_not_leaf: bool
        """
        for file in self.files():
            try:
                leaf = Leaf(file)
                tmp = self._prune_leaf(leaf)
                _ = self._shape(tmp)
            except (AtomError, ValueError, LeafError):
                if delete_not_leaf:
                    logger.warning(f'Delete file that is not leaf: {file}')
                    file.delete()

        levels_and_folders = [(len(x.parts()), x) for x in self.folders()]
        levels_and_folders.sort(reverse=True)
        for _, folder in levels_and_folders:
            if folder.is_empty():
                folder.remove()

    def weeds(self):
        """
        Scans the tree for files with invalid shape, folder or name

        :return: Dictionary of weeds
        :rtype: dict
        """
        # TODO: Move weeds output to a custom class
        # TODO: Refactor to use walk, so that it can handle large trees, and
        # TODO: can be restarted
        leaves = list(self.leaves(ignore_weeds=True))
        files = list(self.files())
        not_leaves = len(files) - len(leaves)
        compression_mismatch = 0
        encryption_mismatch = 0
        for leaf in leaves:
            if leaf.is_compressed() != self.trunk.compress_all:
                compression_mismatch += 1
            if leaf.is_encrypted() != self.trunk.encrypt_all:
                encryption_mismatch += 1
        empty_folders = sum([x.is_empty() for x in self.folders()])
        return {
            'not_leaves': not_leaves,
            'compression_mismatch': compression_mismatch,
            'encryption_mismatch': encryption_mismatch,
            'empty_folders': empty_folders,
            'total_weeds': not_leaves +
                           compression_mismatch +
                           encryption_mismatch +
                           empty_folders
        }

    def total_weeds(self):
        """
        Scan tree and count total number of invalid leaves (weeds)

        :return: Total number of weeds in tree
        :rtype: int
        """
        weeds = self.weeds()
        return weeds['total_weeds']



import logging
import os
import shutil
import glob

from mjooln.filesystem.path import Path

logger = logging.getLogger(__name__)


class FolderError(Exception):
    pass


# TODO: Allow volume as folder
class Folder(Path):

    @classmethod
    def join(cls, *args):
        # Purely cosmetic for IDE
        return super().join(*args)

    @classmethod
    def home(cls):
        """ Get path to user home folder

        Wrapper for ``os.path.expanduser()``

        :return: Path to home folder
        :rtype: Path
        """
        return Folder(os.path.expanduser('~'))

    @classmethod
    def current(cls):
        """ Get current folder path

        Wrapper for ``os.getcwd()``

        :return: Path to current folder
        :rtype: Path
        """
        return cls(os.getcwd())

    def __new__(cls, path, *args, **kwargs):
        self = super(Folder, cls).__new__(cls, path)
        # if cls.RESERVED in instance:
        #     raise FolderError(f'Folder path cannot contain \'{cls.RESERVED}\''
        #                       f'but was found here: {cls.RESERVED}')
        if self.exists():
            if self.is_file():
                raise FolderError(f'Path is a file, '
                                  f'not a folder: {self}')
        return self

    def __repr__(self):
        return f'Folder(\'{self}\')'

    def create(self, error_if_exists=True):
        """ Create new folder, including non existent parent folders

        :raises FolderError: If folder already exists,
            *and* ``error_if_exists=True``
        :param error_if_exists: Error flag. If True, method will raise an
            error if the folder already exists
        :type error_if_exists: bool
        :returns: True if it was created, False if not
        :rtype: bool
        """
        if not self.exists():
            os.makedirs(self)
            return True
        else:
            if error_if_exists:
                raise FolderError(f'Folder already exists: {self}')
            return False

    def touch(self):
        """ Create folder if it does not exist, ignore otherwise

        :return: True if folder was created, False if not
        :rtype: bool
        """
        self.create(error_if_exists=False)

    def untouch(self):
        """ Remove folder if it exists, ignore otherwise
        :return: True if folder was removed, False otherwise
        :rtype: bool
        """
        self.remove(error_if_not_exists=False)

    def parent(self):
        """ Get parent folder

        :return: Parent folder
        :rtype: Folder
        """
        return Folder(os.path.dirname(self))

    def append(self, *args):
        """ Append strings or list of strings to current folder

        Example::

            f = Folder.current()
            print(f)
                '/Users/zaphod'

            f.append('dev', 'code', 'donald')
                '/Users/zaphod/dev/code/donald'

            parts = ['dev', 'code', 'donald']
            f.append(parts)
                '/Users/zaphod/dev/code/donald'

        :param args: Strings or list of strings
        :return: Appended folder as separate object
        :rtype: Folder
        """
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, list):
                return Folder.join(self, '/'.join(arg))
            else:
                return Folder.join(self, arg)
        else:
            return Folder.join(self, *args)

    def is_empty(self):
        """ Check if folder is empty

        :raise FolderError: If folder does not exist
        :return: True if empty, False if not
        :rtype: bool
        """
        if self.exists():
            return len(list(self.list())) == 0
        else:
            raise FolderError(f'Cannot check if non existent folder '
                              f'is empty: {self}')

    # TODO: Add test for empty, with missing name
    def empty(self, name: str):
        """  Recursively deletes all files and subfolders

        .. warning:: Be careful using this, as there is no double check
            before all contents is deleted. Recursively

        :param name: Name of folder. Required to verify deleting all
            contents
        :raise FolderError: If folder does not exist
        :return: Disk usage of folder and subfolders
        :rtype: int
        """
        if self.name() != name:
            raise FolderError(f'Name of folder required to verify '
                              f'deletion: name={self.name()}')
        if self.exists():
            for name in os.listdir(self):
                path = os.path.join(self, name)
                if os.path.isfile(path) or os.path.islink(path):
                    os.unlink(path)
                elif os.path.isdir(path):
                    shutil.rmtree(path)
        else:
            raise FolderError(f'Cannot empty a non existent folder: {self}')

    def remove(self,
               error_if_not_exists: bool = True):
        """ Remove folder

        Will raise an error if not empty

        :raises FolderError: If folder already exists, *and* ``error_if_exists``
            is set to ``True``
        :param error_if_exists: Error flag. If True, method will raise an
            error if the folder already exists
        :type error_if_exists: bool
        """
        if self.exists():
            os.rmdir(self)
        else:
            if error_if_not_exists:
                raise FolderError(f'Cannot remove a non existent '
                                  f'folder: {self}')

    def list(self,
             pattern: str = '*',
             recursive: bool = False):
        if not self.exists():
            raise FolderError(f'Cannot list non existent folder: {self}')

        if recursive:
            paths = glob.glob(self.append('**', pattern),
                              recursive=recursive)
        elif pattern is None:
            paths = os.listdir(self)
        else:
            paths = glob.glob(self.append(pattern))
        return [Path(x) for x in paths]



import json
import simplejson
import yaml as YAML

from mjooln.core.dic import Dic, DicError
from mjooln.core.seed import Seed


class DocError(DicError):
    pass


class JSON:
    """Dict to/from JSON string, with optional human readable"""

    @classmethod
    def dumps(cls, dic, human=True, sort_keys=False, indent=4 * ' '):
        """Convert from dict to JSON string

        :param dic: Input dictionary
        :type dic: dict
        :param human: Human readable flag
        :param sort_keys: Sort key flag (human readable only)
        :param indent: Indent to use (human readable only)
        """
        if human:
            return simplejson.dumps(dic, sort_keys=sort_keys, indent=indent)
        else:
            return json.dumps(dic)

    @classmethod
    def loads(cls, json_string):
        """ Parse JSON string to dictionary

        :param json_string: JSON string
        :return: dict
        """
        return simplejson.loads(json_string)





# TODO: Rewrite as serializer/deserializer
# TODO: In other words, doc will be dic with dates as strings.
# TODO: Atom can be builtin, otherwise it will be a bit tricky.
# TODO: SHould zulu/key/identity also be builtin?
class Doc(Dic):
    """ Enables child classes to mirror attributes, dictionaries and JSON
    strings

    Special cases:

    - Zulu objects will be converted to an ISO 8601 string before a dictionary
      is converted to JSON
    - ISO 8601 strings that are time zone aware with UTC, will be converted to
      Zulu objects after JSON document has been converted to a dictionary
    - Elements that are dictionaries with key names corresponding to Atom
      (key, zulu, identity), will be recognized and converted back to an Atom
      object after JSON document has been converted to a dictionary
    """

    @classmethod
    def from_doc(cls, doc: dict):
        """
        Override in child class to handle non serializable items

        :param doc: Dictionary with serializable items only
        :return: New Doc object instantiated with input dictionary
        :rtype: Doc
        """
        return cls.from_dict(doc)

    @classmethod
    def from_json(cls,
                  json_string: str):
        doc = JSON.loads(json_string=json_string)
        return cls.from_doc(doc)

    @classmethod
    def from_yaml(cls,
                  yaml_string: str):
        doc = YAML.safe_load(yaml_string)
        return cls.from_doc(doc)

    def to_doc(self,
               ignore_private: bool = True):
        # doc = self.to_dict(ignore_private=ignore_private)
        doc = vars(self).copy()
        if ignore_private:
            pop_keys = [x for x in doc
                        if x.startswith(self._PRIVATE_STARTSWITH)]
            for key in pop_keys:
                doc.pop(key)
        for key, item in doc.items():
            if isinstance(item, Doc):
                doc[key] = item.to_doc(ignore_private=ignore_private)
            elif isinstance(item, Seed):
                doc[key] = item.seed()

        return doc

    def to_json(self,
                human_readable: bool = False,
                ignore_private: bool = True):
        doc = self.to_doc(ignore_private=ignore_private)
        return JSON.dumps(doc, human=human_readable)

    def to_yaml(self,
                ignore_private: bool = True):
        doc = self.to_doc(ignore_private=ignore_private)
        return YAML.safe_dump(doc)




import re


class Name:
    """
    Module constants
    """
    #: Minimum length of Elements in Key
    MINIMUM_ELEMENT_LENGTH = 2

    #: Element separator in Key
    ELEMENT_SEPARATOR = '__'

    #: Class separator in Atom
    CLASS_SEPARATOR = '___'

    _CAMEL_TO_SNAKE = r'(?<!^)(?=[A-Z])'
    _SNAKE_TO_CAMEL = r'(.+?)_([a-z])'
    _RE_CAMEL_TO_SNAKE = re.compile(_CAMEL_TO_SNAKE)
    _RE_SNAKE_TO_CAMEL = re.compile(_SNAKE_TO_CAMEL)

    @classmethod
    def camel_to_snake(cls, camel: str):
        """
        Convert camel to snake::

            Name.camel_to_snake('ThisIsCamel')
                this_is_camel
        """
        return cls._RE_CAMEL_TO_SNAKE.sub('_', camel).lower()

    @classmethod
    def snake_to_camel(cls, snake: str):
        """
        Convert snake to camel::

            Name.snake_to_camel('this_is_snake')
                ThisIsSnake


        :param snake:
        :return:
        """
        # TODO: Implement regex instead
        return ''.join(x[0].upper() + x[1:] for x in
                       snake.split('_'))



from functools import lru_cache

from mjooln import Name, Folder, Key, Element, Atom, FileName, File, \
    FileSystem
from mjooln.tree.root import Root
from mjooln.tree.tree import Tree


class GroundProblem(Exception):
    pass


class NoGround(Exception):
    pass


class Ground(Folder):
    """
    Custom hidden folder marking a specific folder as ``Ground``.

    Intended usage is for marking a location in a file system or network drive
    for simplified access. Roots in folder may be accessed directly, and
    attributes and statistics regarding all roots in the underlying folder
    structure may be stored in the ``cave``, which is a root folder hidden
    under ``ground``.
    """

    CAVE = Element('cave')
    GROUND_PREFIX = f'.ground{Name.CLASS_SEPARATOR}'
    CAVE_FOLDER_NAME = CAVE
    CAVE_FILE_NAME = FileName.make(CAVE,
                                   is_hidden=True,
                                   extension=FileName.YAML)

    # @classmethod
    # def _folder(cls, folder):
    #     if isinstance(folder, Folder):
    #         return folder
    #     return Folder(folder)

    @classmethod
    def _key_from_folder(cls, folder):
        return Key(folder.name().replace(cls.GROUND_PREFIX, ''))

    @classmethod
    def _get_ground_folder(cls, folder, key):
        return folder.append(cls.GROUND_PREFIX + key)

    @classmethod
    def _cave_and_file(cls, ground_folder, key):
        cave = Folder.join(ground_folder, cls.CAVE_FOLDER_NAME)
        cave_file = File.join(ground_folder, cls.CAVE_FILE_NAME)
        return cave, cave_file

    @classmethod
    def _get_key(cls, folder):
        ground_folders = FileSystem.list_folders(folder,
                                                 pattern=cls.GROUND_PREFIX +
                                                         '*')
        keys = [cls._key_from_folder(x) for x in ground_folders]
        if len(keys) > 1:
            raise GroundProblem(f'Found multiple grounds in folder: {folder}; '
                                f'Keys: {keys}')
        elif len(keys) == 0:
            raise NoGround(f'No ground found in folder: {folder}')
        return keys[0]

    @classmethod
    def is_in(cls, folder):
        """
        Check if folder is settled ground

        :raise GroundProblem: If folder contains multiple grounds, which is
            is a major problem
        :param folder: Folder to check
        :type folder: Folder
        :return: True if folder contains ground, False if not
        :rtype: bool
        """
        folder = Folder.elf(folder)
        num_paths = len(FileSystem.list_folders(folder,
                                                cls.GROUND_PREFIX + '*',
                                                recursive=False))
        if num_paths == 0:
            return False
        elif num_paths == 1:
            return True
        else:
            raise GroundProblem(f'Folder has multiple grounds: {folder}')

    @classmethod
    def _grounds(cls, folder, depth=0, max_depth=1):
        for fo in FileSystem.list_folders(folder):
            if cls.is_in(fo):
                yield Ground(fo)
            elif depth < max_depth:
                for ground in \
                        cls._grounds(fo, depth=depth + 1, max_depth=max_depth):
                    yield ground

    @classmethod
    def find_ground(cls, folder, max_depth=1):
        folder = Folder.elf(folder)
        grounds = []
        if cls.is_in(folder):
            grounds.append(Ground(folder))
        for ground in cls._grounds(folder, depth=1, max_depth=max_depth):
            grounds.append(ground)
        return grounds

    @classmethod
    def settle(cls, folder, key):
        """
        Create new ground in the specified folder path, with the given key.

        :param folder: Folder to settle
        :type folder: Folder
        :param key: Ground key
        :type key: Key
        :return: Settled ground
        :rtype: Ground
        """
        folder = Folder.elf(folder)
        key = Key.elf(key)
        if not folder.exists():
            raise GroundProblem(f'Cannot settle ground in non existent '
                                f'path: {folder}')

        if cls.is_in(folder):
            raise GroundProblem(f'This folder has already been '
                                f'settled: {folder}; '
                                f'Key: {cls(folder).atom.key()}; '
                                f'Use Ground(folder), not settle()')

        ground_folder = cls._get_ground_folder(folder, key)
        ground_folder.create()
        atom = Atom(key)
        cave, cave_file = cls._cave_and_file(ground_folder, key)
        cave.create()
        cave_file.write(atom)
        return cls(folder)

    @classmethod
    @lru_cache()
    def home(cls):
        home = Folder.home()
        if not Ground.is_in(home):
            return cls.settle(home, key=Key('home'))
        return cls(home)

    def __init__(self, folder):
        super(Ground, self).__init__()
        self.atom = None
        self.__cave = None
        key = self._get_key(folder)
        self._ground_folder = self._get_ground_folder(folder, key)
        self.__cave, cave_file = self._cave_and_file(self._ground_folder, key)
        if not self.__cave.exists():
            raise GroundProblem(f'Missing cave folder: {self.__cave}')
        if not cave_file.exists():
            raise GroundProblem(f'Missing cave file: {cave_file}')
        self.atom = Atom.from_doc(cave_file.read())
        if self.atom.key() != key:
            raise GroundProblem(f'Ground key mismatch. '
                                f'Key in ground folder: {key}; '
                                f'Key in cave file: {self.atom.key()};'
                                f'Cave file: {cave_file}')

    def __repr__(self):
        return f'Ground(\'{self}\')'

    @lru_cache()
    def cave(self):
        return self.__cave

    def walk(self,
             include_files: bool = True,
             include_folders: bool = True):
        for x in FileSystem.walk(Folder(self),
                                 include_files=include_files,
                                 include_folders=include_folders):
            if not x.startswith(self._ground_folder):
                yield x

    def files(self):
        return self.walk(include_files=True,
                         include_folders=False)

    def folders(self):
        return self.walk(include_files=False,
                         include_folders=True)

    def list(self, pattern='*', recursive=False):
        paths = FileSystem.list(Folder(self),
                                pattern=pattern,
                                recursive=recursive)
        return [x for x in paths if not x.startswith(self._ground_folder)]

    def list_files(self, pattern='*', recursive=False):
        paths = self.list(pattern=pattern, recursive=recursive)
        return [x for x in paths if isinstance(x, File)]

    def list_folders(self, pattern='*', recursive=False):
        paths = self.list(pattern=pattern, recursive=recursive)
        return [x for x in paths if isinstance(x, Folder)]

    @classmethod
    def _roots(cls, folder, depth=0, max_depth=1):
        for fo in FileSystem.list_folders(folder):
            if Root.is_root(fo):
                yield Root(fo)
            elif depth < max_depth:
                for root in \
                        cls._roots(fo, depth=depth + 1, max_depth=max_depth):
                    yield root

    def roots(self, max_depth=1):
        # TODO: Rewrite to handle any species of root
        """
        Find all roots recursively from ground folder

        :return: List of all roots
        :rtype: [Root]
        """
        roots = []
        for r in self._roots(self, depth=1, max_depth=max_depth):
            roots.append(r)
        return roots

    @classmethod
    def _trees(cls, folder, depth=0, max_depth=1):
        for fo in FileSystem.list_folders(folder):
            if Tree.is_tree(fo):
                yield Tree(fo)
            elif depth < max_depth:
                for tree in \
                        cls._trees(fo, depth=depth + 1, max_depth=max_depth):
                    yield tree

    def trees(self, max_depth=1):
        # TODO: Somehow merge roots and trees
        """
        Find all roots recursively from ground folder

        :return: List of all roots
        :rtype: [Root]
        """
        trees = []
        for r in self._trees(self, depth=1, max_depth=max_depth):
            trees.append(r)
        return trees

    # def unsettle(self, force=False, key=None):
    #     if not self.__cave.is_empty():
    #         if not force or key != self.atom.key():
    #             raise GroundProblem(f'Cave is not empty. Use force=True, and '
    #                                 f'key={self.atom.key()}')
    #         self.__cave.empty()
    #     self.__cave.remove()
    #     _, cave_file = self._cave_and_file(self._ground_folder,
    #                                        self.atom.key())
    #     cave_file.delete()
    #     self.remove()
